from schematics.models import Model
from schematics.types import StringType, IntType

from .. import api, entities as en, models as m
from ..enums import *
from ..error import BoardError
from ..models import MondayModel


class _Board(Model):
    """The base data model for a board"""

    id = StringType(required=True)
    name = StringType()
    board_folder_id = IntType()
    board_kind = StringType()
    communication = StringType()
    description = StringType()
    permissions = StringType()
    pos = StringType()
    state = StringType()
    workspace_id = StringType()
    updated_at = StringType()


class Board(_Board):
    """The entity model for a board

        Properties

            activity_logs : `list[moncli.entities.ActivityLog]`
                The board log events.
            board_folder_id : `int`
                The board's folder unique identifier.
            board_kind : `str`
                The board's kind (public / private / share).
            columns : `moncli.entities.objects.ColumnCollection`
                The board's visible columns.
            communication : `str`
                Get the board communication value - typically meeting ID.
            description : `str`
                The board's description.
            groups : `list[moncli.entities.Group]`
                The board's visible groups.
            id : `str`
                The unique identifier of the board.
            items : `list[moncli.entities.Item]`
                The board's items (rows).
            name : `str`
                The board's name.
            owner : `moncli.entities.User`
                The owner of the board.
            permissions : `str`
                The board's permissions.
            pos : `str`
                The board's position.
            state : `str`
                The board's state (all / active / archived / deleted).
            subscribers : `list[moncli.entities.User]`
                The board's subscribers.
            tags : `list[moncli.entities.objects.Tag]`
                The board's specific tags.
            top_group : `moncli.entities.Group`
                The top group at this board.
            updated_at : `str`
                The last time the board was updated at (ISO8601 DateTime).
            updates : `list[moncli.entities.Update]`
                The board's updates.
            views : `list[moncli.entities.BoardView]`
                The board's views.
            workspace : `moncli.entities.Workspace`
                The workspace that contains this board (null for main workspace).
            workspace_id : `str`
                The board's workspace unique identifier (null for main workspace).

        Methods

            get_activity_logs : `list[moncli.entities.ActivityLog]`
                Get the board log events. 
            get_views : `list[moncli.entities.BoardView]`
                Get the board's views.
            add_subscribers : `list[moncli.entities.User]`
                Add subscribers to this board.
            get_subscribers : `list[monlci.entities.User]`
                Get board subscribers.
            delete_subscribers : `list[monlci.entities.User]`
                Remove subscribers from the board.
            add_column : `moncli.entities.Column`
                Create a new column in board.
            change_column_title : 'moncli.entities.board'
                Change title of a column.
            get_columns : `list[moncli.entities.Column]`
                Get the board's visible columns.
            add_group : `moncli.entities.Group`
                Creates a new group in the board.
            get_groups : `list[moncli.entities.Group]`
                Get the board's visible groups.
            get_group : `moncli.entities.Group`
                Get a group belonging to the board by ID or title.
            add_item : `moncli.entities.Item`
                Create a new item in the board.
            get_items : `list[moncli.entities.Item]`
                Get the board's items (rows).
            get_items_by_column_values : `list[moncli.entities.Item]`
                Search items in this board by their column values.
            get_column_value : `moncli.entities.ColumnValue`
                Create a column value from a board's column.
            create_webhook : `moncli.entities.Webhook`
                Create a new webhook.
            delete_webhook : `moncli.entities.Webhook`
                Delete a new webhook.
            get_workspace : `moncli.entities.Workspace`
                Get the board's workspace that contains this board (null for main workspace).
            get_updates : `list[moncli.entities.Update]`
                Get the board's updates.
            get_tags : `list[moncli.entities.Tag]`
                Get the board's tags.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds', None)
        self.__activity_logs = kwargs.pop('__activity_logs', None)
        self.__columns = en.BaseColumnCollection()
        self.__groups = kwargs.pop('__groups', None)
        self.__items = kwargs.pop('__items', None)
        self.__subscribers = kwargs.pop('__subscribers', None)
        self.__views = kwargs.pop('__views', None)
        self.__tags = kwargs.pop('__tags', None)
        self.__updates = kwargs.pop('_updates', None)
        self.__workspace = kwargs.pop('__workspace', None)

        activity_logs = kwargs.pop('activity_logs', None)
        columns = kwargs.pop('columns', None)
        groups = kwargs.pop('groups', None)
        items = kwargs.pop('items', None)
        subscribers = kwargs.pop('subscribers', None)
        tags = kwargs.pop('tags', None)
        updates = kwargs.pop('updates', None)
        views = kwargs.pop('views', None)
        workspace = kwargs.pop('workspace', None)

        super(Board, self).__init__(kwargs)
        
        if activity_logs and not self.__activity_logs:
            self.__activity_logs = [en.ActivityLog(log) for log in activity_logs]
        if columns and not self.__columns:
            self.__columns = en.BaseColumnCollection([en.Column(**column) for column in columns])
        if groups and not self.__groups:
            self.__groups = [en.Group(creds=self.__creds, **group) for group in groups]
        if items and not self.__items:
            self.__items = [en.Item(creds=self.__creds, __board=self, **item) for item in items]
        if subscribers and not self.__subscribers:
            self.__items = [en.User(creds=self.__creds, **subscriber) for subscriber in subscribers]
        if tags and not self.__tags:
            self.__tags = [en.Tag(tag) for tag in tags]
        if views and not self.__views:
            self.__views = [en.BoardView(view) for view in views]
        if updates and not self.__updates:
            self.__updates = [en.Update(update) for update in updates]
        if workspace and not self.__workspace:
            self.__workspace = en.Workspace(workspace)
        

    def to_primitive(self, role=None, app_data=None, **kwargs):
        o = super().to_primitive(role, app_data, **kwargs)
        if self.__activity_logs != None:
            o['activity_logs'] = [log.to_primitive() for log in self.__activity_logs]
        if self.__columns:
            o['columns'] = [column.to_primitive() for column in self.__columns]
        if self.__groups:
            o['groups'] = [group.to_primitive() for group in self.__groups]
        if self.__items:
            o['items'] = [item.to_primitive() for item in self.__items]
        if self.__items:
            o['subscribers'] = [item.to_primitive() for item in self.__subscribers]
        if self.__tags:
            o['tags'] = [item.to_primitive() for item in self.__tags]
        if self.__updates:
            o['updates'] = [item.to_primitive() for item in self.__updates]
        if self.__items:
            o['views'] = [item.to_primitive() for item in self.__views]
        return o

    def __repr__(self):
        return str(self.to_primitive())
 
    @property
    def activity_logs(self):
        """The board log events"""

        if self.__activity_logs == None:
            self.__activity_logs = self.get_activity_logs()
        return self.__activity_logs

    @property
    def columns(self):
        """Retrieve board columns"""

        if not self.__columns:
            self.__columns = self.get_columns()
        return self.__columns

    @property
    def groups(self):
        """Retrieve board groups"""
        
        if not self.__groups:
            self.__groups = self.get_groups()
        return self.__groups

    @property
    def items(self):
        """Retrieve board items"""

        if not self.__items:
            self.__items = self.get_items()
        return self.__items

    @property
    def subscribers(self):
        """Retrieve board subscribing users"""

        if not self.__subscribers:
            self.__subscribers = self.get_subscribers()
        return self.__subscribers

    @property
    def tags(self):
        """Retrieve board tags."""
        if self.__tags == None: 
            self.__tags = self.get_tags()
        return self.__tags

    @property
    def updates(self):
        """Retrieve board updates."""
        if self.__updates == None: 
            self.__updates = self.get_updates()
        return self.__updates

    @property
    def views(self):
        """Retrieve board updates."""
        if self.__views == None:
            self.__views = self.get_views()
        return self.__views

    @property
    def workspace(self):
        """Retrieve workspace"""

        if not self.__workspace:
            self.__workspace = self.get_workspace()
        return self.__workspace


    def get_activity_logs(self, *args, **kwargs):
        """Get board log events.
        
            Parameters
            
                args : `tuple`
                    The list of activity log return fields.
                kwargs : `dict`
                    Optional keyword arguments for retrieving activity logs.

            Returns

                activity_logs : `list[moncli.entities.ActivityLog]`
                    The board's activity logs.

            Return Fields

                account_id : `str`
                    The unique identifier of the user's account.
                created_at : `str`
                    The create date
                data : `str`
                    The item's column values in string form.
                entity : `str`
                    The activity log's entity.
                event : `str`
                    The activity log's event.
                id : `str`
                    The activity log's unique identifier.
                user_id : `str`
                    The user's unique identifier.

            Optional Arguments

                limit : `int`
                    Number of items to get, the default is 25.
                page : `int`
                    Page number to get, starting at 1.
                user_ids : `list[str]`
                    User ids to filter.
                column_ids : `list[str]`
                    Column ids to filter.
                group_ids : `list[str]`
                    Group ids to filter.
                item_ids : `list[str]`
                    Item id to filter
                from : `str`
                    From timespamp (ISO8601).
                to : `str`
                    To timespamp (ISO8601).
        """

        if kwargs:
            kwargs = {'activity_logs': kwargs}

        activity_logs_data = api.get_boards(
            'id', 'name',
            *api.get_field_list(api.DEFAULT_ACTIVITY_LOG_QUERY_FIELDS, 'activity_logs', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id],
            **kwargs)[0]['activity_logs']

        return [en.ActivityLog(activity_log) for activity_log in activity_logs_data]


    def get_views(self, *args, **kwargs):
        """Get the board's views.

            Parameters

                args : `tuple`
                    The list of board view return fields.
                kwargs : `dict`
                    Optional keyword arguments for querying board views.

            Returns

                views : `list[moncli.entities.BoardView]`
                    The board's collection of board views.

            Return Fields

                id : `str`
                    The view's unique identifier.
                name : `str`
                    The view's name.
                settings_str : `str`
                    The view's settings in a string from.
                type : `str`
                    The view's type.

            Optional Arguments

                ids : `str`
                    The list of view unique identifiers.
                type :  `str`
                    The view's type.
        """

        if kwargs:
            kwargs = {'views': kwargs}

        views_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_BOARD_VIEW_QUERY_FIELDS, 'views', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id],
            **kwargs)[0]['views']

        return [en.BoardView(view) for view in views_data]


    def add_subscribers(self, user_ids: list, *args, **kwargs):
        """Add subscribers to this board.

            Parameters

                user_ids : `list[str]`
                    User ids to subscribe to a board.
                args : `tuple`
                    List of user return fields.
                kwargs : `dict`
                    Additional keyword arguments for adding subscribers to board.
            
            Returns

                user : `list[moncli.entity.User]`
                    The users subscribed to this board.
            
            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_verified: `bool`
                    Is the user is verified
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.

            Optional Arguments

                kind : `moncli.enums.SubscriberKind`
                    Subscribers kind (subscriber / owner).
        """

        subscribers_data = api.add_subscribers_to_board(
            self.id,
            user_ids,
            *args,
            api_key=self.__creds.api_key_v2,
            **kwargs)

        return [en.User(creds=self.__creds, **user) for user in subscribers_data]

    
    def get_subscribers(self, *args):
        """Get board subscribers

            Parameters

                args : `tuple`
                    The list of user return fields.
                
            Returns

                subscribers : `list[moncli.entity.User]`
                    The board subscribers.

            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_verified: `bool`
                    Is the user is verified
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """

        kwargs = {
            'ids': [self.id]
        }

        users_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_USER_QUERY_FIELDS, 'subscribers', *args),
            api_key=self.__creds.api_key_v2,
            **kwargs)[0]['subscribers']

        return [en.User(creds=self.__creds, **user) for user in users_data]


    def delete_subscribers(self, user_ids: list, *args):
        """Remove subscribers from the board.

            Parameters

                user_ids : `list[str]`
                    User ids to unsubscribe from board.
                args : `tuple`
                    The list of user return fields.

            Returns

                subscribers : `list[moncli.entities.User]`
                    The users unsubscribed from the board.
            
            Return Fields
            
                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_verified: `bool`
                    Is the user is verified
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """
        
        users_data = api.delete_subscribers_from_board(
            self.id,
            user_ids,
            *args,
            api_key=self.__creds.api_key_v2)

        return [en.User(creds=self.__creds, **user) for user in users_data]


    def add_column(self, title: str, column_type: ColumnType, *args, **kwargs):
        """Create a new column in board.
    
            Parameters

                title : `str`
                    The new column's title.
                column_type : `moncli.enums.ColumnType`
                    The type of column to create.
                args : `tuple`
                    The list of column return fields.
                kwargs : `dict`
                    The optional keywork arguments.

            Returns

                column : `moncli.entities.Column`
                    The created column.

            Return Fields

                archived : `bool`
                    Is the column archived or not.
                id : `str`
                    The column's unique identifier.
                pos : `str`
                    The column's position in the board.
                settings_str : `str`
                    The column's settings in a string form.
                title : `str`
                    The column's title.
                type : `str`
                    The column's type.
                width : `int`
                    The column's width.

            Optional Arguments

                defaults : `json`
                    The new column's defaults.
        """

        column_data = api.create_column(
            self.id, 
            title, 
            column_type,  
            *args,
            api_key=self.__creds.api_key_v2,
            **kwargs)

        column = en.Column(**column_data)
        self.__columns.append(column)
        return column

    

    def change_column_title(self,  title: 'str', column, *args):
        """Change title of a column.

           Parameters

                column : 'str'
                    The column object to be updated.
                title : 'str'
                    The new title of the column.
                args : 'tuple'
                    The collection of item return fields.

            Returns

                column :  'moncli.entities.column.Column'
                    The updated column title.

            Return Fields

                archived : `bool`
                    Is the column archived or not.
                id : `str`
                    The column's unique identifier.
                pos : `str`
                    The column's position in the board.
                settings_str : `str`
                    The column's settings in a string form.
                title : `str`
                    The column's title.
                type : `str`
                    The column's type.
                width : `int`
                    The column's width.
        """
        
        column_data = api.change_column_title(
            title,
            column.id,
            self.id,
            *args,
            api_key=self.__creds.api_key_v2)
        
        return en.Column(**column_data)

   
    def get_columns(self, *args, **kwargs):
        """Get the board's visible columns.

            Parameters

                args : `tuple`
                    The list of column return fields.
                kwargs : `dict`
                    The optional keywork arguments.

            Returns

                columns : `list[moncli.entities.Column]`
                    The board's columns.

            Return Fields

                archived : `bool`
                    Is the column archived or not.
                id : `str`
                    The column's unique identifier.
                pos : `str`
                    The column's position in the board.
                settings_str : `str`
                    The column's settings in a string form.
                title : `str`
                    The column's title.
                type : `str`
                    The column's type.
                width : `int`
                    The column's width.

            Optional Arguments

                ids : `str`
                    A list of column unique identifiers.
        """

        column_kwargs = {}
        if kwargs:
            column_kwargs['columns'] = kwargs

        column_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_COLUMN_QUERY_FIELDS, 'columns', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)],
            limit=1,
            **column_kwargs)[0]['columns']

        return en.BaseColumnCollection([en.Column(**data) for data in column_data])


    def add_group(self, group_name: str, *args):
        """Creates a new group in the board.
    
            Parameters

                group_name : `str`
                    The name of the new group.
                args : `tuple`
                    The list of group return fields.

            Returns

                group : `moncli.entities.Group`
                    The created group.

            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
        """

        group_data = api.create_group(
            self.id,
            group_name,
            *args,
            api_key=self.__creds.api_key_v2)

        return en.Group(
            creds=self.__creds,
            __board=self,
            **group_data)


    def get_groups(self, *args, **kwargs):
        """Get the board's visible groups.

            Parameters

                args : `tuple`
                    The list of group return fields.
                kwargs : `dict`
                    Optional keyword arguments for getting board groups.

            Returns

                groups : `list[moncli.entities.Groups]`
                    The board's groups.

            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.

            Optional Arguments

                ids : `list[string]`
                    A list of group unique identifiers.
        """

        group_kwargs = {}
        if kwargs:
            group_kwargs['groups'] = kwargs

        groups_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_GROUP_QUERY_FIELDS, 'groups', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)],
            **group_kwargs)[0]['groups']

        return [en.Group(creds=self.__creds, __board=self, **data) for data in groups_data]
  

    def get_group(self, id: str = None, title: str = None, *args):
        """Get a group belonging to the board by ID or title.

            Parameters

                id : `str`
                    The group's unique identifier.
                    NOTE: This parameter is mutually exclusive and cannot be used with 'title'.
                title : `str`
                    The group's title.
                    NOTE: This parameter is mutually exclusive and cannot be used with 'id'.
                args : `tuple`
                    The list of group return fields.
        
            Returns

                group : `moncli.entities.Group`
                    The board's requested group.
        
            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
        """

        if id is None and title is None:
            raise NotEnoughGetGroupParameters()
        
        if id is not None and title is not None:
            raise TooManyGetGroupParameters()

        if id is not None:
            return self.get_groups(*args, ids=[id])[0]
        
        else:
            return [group for group in self.get_groups(*args) if group.title == title][0]


    def add_item(self, item_name: str,as_model: type = None, get_column_values = False, *args, **kwargs):
        """Create a new item in the board.

            Parameters

                item_name : `str`
                    The new item's name.
                as_model: `type`
                    The MondayModel subclass to be returned.
                get_column_values: `bool`
                    Returns column values with created item if set to `True`.


            Returns

                item : `moncli.entities.Item`
                    The created item.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.

            Optional Arguments

                group_id : `str`
                    The group's unique identifier.
                column_values : `json`
                    The column values of the new item.
        """

        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])

        column_values = kwargs.pop('column_values', None)
        if column_values:
            if type(column_values) == dict:
                kwargs['column_values'] = column_values
            elif type(column_values) == list:
                kwargs['column_values'] = { value.id: value.format() for value in column_values }
            else:
                raise InvalidColumnValue(type(column_values).__name__)

        item_data = api.create_item(
            item_name, 
            self.id, 
            *args, 
            api_key=self.__creds.api_key_v2, 
            **kwargs)

        items =  en.Item(creds=self.__creds, __board=self, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise BoardError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


    def get_items(self, get_column_values: bool = False, as_model: type = None, *args, **kwargs):
        """Get the board's items (rows).

            Parameters

                args : `tuple`
                    The list of item return fields.
                get_column_values: `bool`
                    Returns column values with items if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                kwargs : `dict`
                    The optional keyword arguments for getting items.

            Returns

                items : `list[moncli.entities.Item]`
                    The board's items.
            
            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.

            Optional Arguments

                ids : `list[str]`
                    The list of items unique identifiers.
                limit : `int`
                    Number of items to get.
                page : `int`
                    Page number to get, starting at 1.
        """
        
        if get_column_values:
            args = list(args)
            for arg in ['items.column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['items.id', 'items.name'])
        else:
            args = api.get_field_list(api.DEFAULT_ITEM_QUERY_FIELDS, 'items', *args)

        item_kwargs = {}
        if kwargs:
            item_kwargs['items'] = kwargs

        items_data = api.get_boards(
            *args, 
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)],
            **item_kwargs)[0]['items']

        items = [en.Item(creds=self.__creds, **item_data) for item_data in items_data] 
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise BoardError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]
 


    def get_items_by_column_values(self, column_value: en.cv.ColumnValue, get_column_values: bool = False, as_model: type = None, *args, **kwargs):
        """Search items in this board by their column values.
    
            Parameters

                column_value : `moncli.entites.ColumnValue`
                    The column value to search on.
                get_column_values: `bool`
                    Flag used to include column values with the returned items.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item return fields.
                kwargs : `dict`
                    The optional keyword arguments for searching items.
        
            Returns

                items : `list[moncli.entities.Item]`
                    The board's queried items.
        
            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments

                limit : `int`
                    Number of items to get.
                page : `int`
                    Page number to get, starting at 1.
                column_id : `str`
                    The column's unique identifier.
                column_value : `str`
                    The column value to search items by.
                column_type : `str`
                    The column type.
                state : `moncli.enumns.State`
                    The state of the item (all / active / archived / deleted), the default is active.
        """
        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])

        if isinstance(column_value, en.cv.DateValue):
            value = column_value.date
        elif isinstance(column_value, en.cv.StatusValue):
            value = column_value.label
        else:
            value = column_value.format()

        items_data = api.get_items_by_column_values(
            self.id, 
            column_value.id, 
            value, 
            *args,
            api_key=self.__creds.api_key_v2, 
            **kwargs)

        items = [en.Item(creds=self.__creds, **item_data) for item_data in items_data] 
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise BoardError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


    
    def get_items_by_multiple_column_values(self, column: en.Column, column_values: list, get_column_values: bool = False, as_model: type = None, *args, **kwargs):
        """Search items in this board by their column values.
    
            Parameters

                column : `moncli.entites.objects.Column`
                    The column to search on.
                column_values : `list[str]`
                    The list of values to search on.
                get_column_values: `bool`
                    Retrieves all item column values if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item return fields.
                kwargs : `dict`
                    The optional keyword arguments for searching items.
        
            Returns

                items : `list[moncli.entities.Item]`
                    The board's queried items.
        
            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments

                limit : `int`
                    Number of items to get.
                page : `int`
                    Page number to get, starting at 1.
                column_id : `str`
                    The column's unique identifier.
                column_value : `str`
                    The column value to search items by.
                column_type : `str`
                    The column type.
                state : `moncli.enumns.State`
                    The state of the item (all / active / archived / deleted), the default is active.
        """
        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])

        if column.column_type == ColumnType.numbers:
            value = [str(value) for value in column_values]
        else:
            value = column_values

        items_data = api.get_items_by_multiple_column_values(
            self.id, 
            column.id, 
            value, 
            *args,
            api_key=self.__creds.api_key_v2, 
            **kwargs)

        items = [en.Item(creds=self.__creds, **item_data) for item_data in items_data] 
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise BoardError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


    def get_column_values(self):
        """This method has not yet been implemented."""
        pass


    def get_column_value(self, id: str = None, title: str = None, **kwargs):
        """Create a column value from a board's column.

            Parameters

                id : `str`
                    The column's unique identifier.
                title : `str`
                    The column's title.
                kwargs : `dict`
                    The raw column value data to add.
        
            Returns

                column_value : `list[moncli.entities.ColumnValue]`
                    A new column_value instance.
        """

        if id is None and title is None:
            raise NotEnoughGetColumnValueParameters()

        if id is not None and title is not None:
            raise TooManyGetColumnValueParameters()

        columns = { column.id: column for column in self.columns }
        if id is not None:
            column = columns[id]
        elif title is not None:
            column = [column for column in columns.values() if column.title == title][0]

        column_type = column.column_type      
        return en.cv.create_column_value(column_type, id=column.id, title=column.title, settings_str=column.settings_str, **kwargs)


    def create_webhook(self, url: str, event: WebhookEventType, *args, **kwargs):
        """Create a new webhook.
    
            Parameters

                url : `str`
                    The webhook URL.
                event : `moncli.enums.WebhookEventType`
                    The event to listen to (incoming_notification / change_column_value / change_specific_column_value / create_item / create_update).
                args : `tuple`
                    The list of webhook return fields.
                kwargs : `dict`
                    The optional keyword arguments for creating a webhook.
        
            Returns

                webhook : `moncli.entities.Webhook`
                    The created webhook.
        
            Return Fields

                board_id : `str`
                    The webhook's board id.
                id : `str`
                    The webhook's unique identifier.
            
            Optional Arguments

                config : `dict`
                    The webhook config.
                    Example: This argument is currenlty only available for the 'change_specific_column_value' event.
                    >>> board.create_webhook('http://test.website.com/webhook/test', WebhookEventType.change_specific_column_value, {'columnId': 'column_1'})
        """

        # Modify kwargs to config if supplied.
        if kwargs:
            if event != WebhookEventType.change_specific_column_value:
                raise WebhookConfigurationError(event)
            kwargs = {'config': kwargs}

        webhook_data = api.create_webhook(
            self.id, 
            url, 
            event,
            *args,
            api_key=self.__creds.api_key_v2, 
            **kwargs)

        webhook_data['is_active'] = True
        return en.Webhook(webhook_data)


    def delete_webhook(self, webhook_id: str, *args):
        """Delete a new webhook.
    
            Parameters

                id : `str`
                    The webhook's unique identifier.
                args : `tuple`
                    The list of webhook return fields.
        
            Returns

                webhook : `moncli.entities.Webhook`
                    The deleted webhook.
        
            Return Fields

                board_id : `str`
                    The webhook's board id.
                id : `str`
                    The webhook's unique identifier.
        """

        webhook_data = api.delete_webhook(
            webhook_id,
            *args, 
            api_key=self.__creds.api_key_v2)

        webhook_data['is_active'] = False
        return en.Webhook(webhook_data)


    def get_workspace(self, *args):
        """Retrieves the board workspace
    
            Parameters

                args : `tuple`
                    The workspace return fields.
        
            Returns

                workspace : `list[moncli.entities.Workspace]`
                    The board workspace.
        
            Return Fields

                id : `str`
                    The workspace's unique identifier.
                name : `str`
                    The workspace's name.
                kind : `str`
                    The workspace's kind (open / closed)
                description : `str`
                    The workspace's description
        """

        workspace_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_WORKSPACE_QUERY_FIELDS, 'workspace', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id])[0]['workspace']

        return en.Workspace(workspace_data)

    
    def get_updates(self, *args, **kwargs):
        """Get the board's updates
 
            Parameters
    
                args : `tuple`
                    Optional update return fields.
                kwargs : `dict`
                    Optional keyword arguments for getting item updates.

            Returns

                update : `list[moncli.entities.Update]`
                    The item's updates.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.entities.Reply]`
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.

            Optional Arguments

                limit : `int`
                    Number of updates to get; the default is 25.
                page : `int`
                    Page number to get, starting at 1.
        """

        updates_kwargs = {}
        if kwargs:
            updates_kwargs = {'updates': kwargs}

        updates_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_UPDATE_QUERY_FIELDS, 'updates', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id],
            **updates_kwargs)[0]['updates']

        return [en.Update(creds=self.__creds, **data) for data in updates_data]

    
    def get_tags(self, *args):
        """Get the board's tags

            Parameters

                args : `tuple`
                    The list of tag return fields.
                kwargs : `dict`
                    Optional keyword arguments for querying tags.

            Returns

                tags : `list[moncli.entities.Tag]`
                    A collection of tags.

            Return Fields

                color : `str`
                    The tag's color.
                id : `str`
                    The tag's unique identifier.
                name : `str`
                    The tag's name.

            Optional Arguments

                ids : `list[str]`
                    The list of tags unique identifiers.
        """

        tags_data = api.get_boards(
            *api.get_field_list(api.DEFAULT_TAG_QUERY_FIELDS, 'tags', *args),
            api_key=self.__creds.api_key_v2,
            ids=[self.id])[0]['tags']

        return [en.Tag(data) for data in tags_data]

        
class TooManyGetGroupParameters(Exception):
    def __init__(self):
        self.message = "Unable to use both 'id' and 'title' when querying for a group."
        

class NotEnoughGetGroupParameters(Exception):
    def __init__(self):
        self.message = "Either the 'id' or 'title' is required when querying a group."


class InvalidColumnValue(Exception):
    def __init__(self, column_value_type: str):
        self.message = "Unable to use column value of type '{}' with the given set of input parameters".format(column_value_type)


class TooManyGetColumnValueParameters(Exception):
    def __init__(self):
        self.message = "Unable to use both 'id' and 'title' when querying for a column value."
        

class NotEnoughGetColumnValueParameters(Exception):
    def __init__(self):
        self.message = "Either the 'id' or 'title' is required when querying a column value."

class WebhookConfigurationError(Exception):
    def __init__(self, event: WebhookEventType):
        self.message = "Webhook event type '{}' does not support configuraitons".format(event.name)



import json

from schematics.models import Model
from schematics.types import StringType

from .. import api, entities as en, models as m, error as e
from ..error import ItemError
from ..models import MondayModel


class _Item(Model):
    """Item Base Model"""

    id = StringType(required=True)
    name = StringType()
    created_at = StringType()
    creator_id = StringType()
    state = StringType()
    updated_at = StringType()


class Item(_Item):
    """An item (table row)
    
        Properties

            assets : `list[moncli.entities.Asset]`
                The item's assets/files.
            board : `moncli.entities.Board`
                The board that contains this item.
            column_values : `list[moncli.entities.ColumnValue]`
                The item's column values.
            created_at : `str`
                The item's create date.
            creator : `moncli.entities.User`
                The item's creator.
            creator_id : `str`
                The item's unique identifier.
            group : `moncli.entities.Group`
                The group that contains this item.
            id : `str`
                The item's unique identifier.
            name : `str`
                The item's name.
            state : `str`
                The board's state (all / active / archived / deleted)
            subitems: list[moncli.entities.Item]
                The item's nested subitems.
            subscriber : `moncli.entities.User`
                The pulse's subscribers.
            updated_at : `str`
                The item's last update date.
            updates : `moncli.entities.Update`
                The item's updates.

        Methods

            add_file : `moncli.entities.Asset`
                Add a file to a column value.
            get_files : `list[moncli.entities.Asset]`
                Retrieves the file assets for the login user's account.
            remove_files : `moncli.entities.Item`
                Removes files from a column value.
            get_board : `moncli.entities.Board`
                Get the board that contains this item.
            get_group : `moncli.entities.Group`
                Get the group containing this item.
            get_creator : `moncli.entities.User`
                Get the item's creator.
            get_column_values : `list[moncli.entities.ColumnValue]`
                Get the item's column values.
            get_column_value : `moncli.entities.ColumnValue`
                Get an item's column value by ID or title.
            get_subitems: list[moncli.entities.Item]
                Get the item's nested subitems
            change_item_name: `moncli.entities.Item`
                change an Item's column name
            change_column_value : `moncli.entities.Item`
                Change an item's column value.
            change_multiple_column_values : `moncli.entities.Item`
                Change the item's column values.
            create_subitem : `moncli.entities.Item`
                Create subitem.
            move_to_group : `moncli.entities.Item`
                Move item to a different group.
            archive : `moncli.entities.Item`
                Archive this item.
            delete : `moncli.entities.Item`
                Delete this item.
            duplicate : `moncli.entities.Item`
                Duplicate this item.
            add_update : `moncli.entities.Update`
                Create a new update for this item.
            get_updates : `list[moncli.entities.Update]`
                Get updates for this item.
            delete_update : `moncli.entities.Update`
                Delete item update.
            clear_updates : `moncli.entities.Item`
                Clear all updates for item.
    """

    def __init__(self, **kwargs):
        self.__creds = kwargs.pop('creds')
        self.__assets = kwargs.pop('__assets', None)
        self.__board = kwargs.pop('__board', None)
        self.__group = kwargs.pop('__group', None)
        self.__creator = kwargs.pop('__creator', None)
        self.__column_values = en.BaseColumnCollection()
        self.__updates = kwargs.pop('__updates', None)
        self. __parent_item = kwargs.pop('__parent_item',None)
        self. __subitems = kwargs.pop('__subitems',None)

        assets = kwargs.pop('assets', None)
        board = kwargs.pop('board', None)
        group = kwargs.pop('group', None)
        creator = kwargs.pop('creator', None)
        column_values = kwargs.pop('column_values', None)
        updates = kwargs.pop('updates', None)
        parent_item = kwargs.pop('parent_item',None)
        subitems = kwargs.pop('subitems',None)

        super(Item, self).__init__(kwargs)

        if assets != None and not self.__assets:
            self.__assets = [en.Asset(creds=self.__creds, **asset) for asset in assets]
        if board and not self.__board:
            self.__board = en.Board(creds=self.__creds, **board)
        if group and not self.__group:
            self.__group = en.Group(creds=self.__creds, board_id=self.board.id, **group)
        if creator and not self.__creator:
            self.__creator = en.User(creds=self.__creds, **creator)
        if column_values != None and len(self.__column_values) == 0:
            #Column values are a process around here...
            columns_map = { column.id: column for column in self.board.columns }
            for data in column_values:
                column = columns_map[data['id']]
                self.__column_values.append(en.cv.create_column_value(column.column_type, settings_str=column.settings_str, **data))
        if updates != None and not self.__updates:
            self.__updates = [en.Update(creds=self.__creds, **update_data) for update_data in updates]
        if parent_item and not self.__parent_item: 
            self.__parent_item = en.Item(creds=self.__creds, **parent_item)
        if subitems and not self.__subitems:
            self.__subitems = [en.Item(creds = self.__creds, **value) for value in subitems]



    def __repr__(self):
        o = self.to_primitive()
        if self.__assets:
            o['assets'] = self.__assets.to_primitive()
        if self.__board:
            o['board'] = self.__board.to_primitive()
        if self.__creator:
            o['creator'] = self.__creator.to_primitive()
        if self.__column_values:
            o['column_values'] = []
            for value in self.__column_values:
                o['column_values'].append(value.to_primitive())
        if self.__updates:
            o['updates'] = [value.to_primitive() for value in self.__updates]
        if self.__parent_item:
            o['parent_item'] = self.__parent_item.to_primitive
        if self.__subitems:
            o['subitems'] = [value.to_primitive() for value in self.__subitems]

        return str(o)

    @property
    def assets(self):
        """The item's assets/files."""
        if self.__assets == None:
            self.__assets = self.get_files()
        return self.__assets

    @property
    def board(self):
        """The board that contains this item."""
        if not self.__board:
            self.__board = self.get_board()
        return self.__board

    @property
    def group(self):
        """The group that contains this item."""
        if not self.__group:
            self.__group = self.get_group()
        return self.__group

    @property
    def creator(self):
        """The item's creator."""
        if not self.__creator:
            self.__creator = self.get_creator()
        return self.__creator

    @property
    def column_values(self):
        """The item's column_values."""
        if len(self.__column_values) == 0:
            self.__column_values = self.get_column_values()
        return self.__column_values

    @property
    def updates(self):
        """The item's updates."""
        if self.__updates == None: 
            self.__updates = self.get_updates()
        return self.__updates
    
    @property
    def parent_item(self):
        """The parent item ."""
        if self.__parent_item == None: 
            self.__parent_item = self.get_parent_item()
        return self.__parent_item

    @property
    def subitems(self):
        """The nested subitems."""
        if not self.__subitems:
            self.__subitems = self.get_subitems()
        return self.__subitems


    def add_file(self, file_column: en.cv.FileValue, file_path: str, *args):
        """Add a file to a column value.

            Parameters

                file_column : moncli.entities.FileValue
                    The file column value to be updated.
                file_path : `str`
                    The file path.
                args : `tuple`
                    Optional file return fields.

            Returns

                assets : `moncli.entities.Asset`
                    The newly created file asset.

            Return Fields

                created_at : `str`
                    The file's creation date.
                file_extension : `str`
                    The file's extension.
                file_size : `int`
                    The file's size in bytes.
                id : `str`
                    The file's unique identifier.
                name : `str`
                    The file's name.
                public_url : `str`
                    Public url to the asset, valid for 1 hour.
                uploaded_by : `moncli.entities.user.User`
                    The user who uploaded the file
                url : `str`
                    The user who uploaded the file
                url_thumbnail : `str`
                    Url to view the asset in thumbnail mode. Only available for images.
        """

        asset_data = api.add_file_to_column(
            self.id,
            file_column.id,
            file_path,
            *args,
            api_key=self.__creds.api_key_v2)
        return en.Asset(**asset_data)


    def get_files(self, column_ids: list = None, *args):
        """Retrieves the file assets for the login user's account.

            Parameters

                args : `str`
                    The list asset return fields.
                kwargs : `dict`
                    Optional keyword arguments for retrieving file assets from an item.
                
            Returns

                assets : `list[moncli.entities.asset.Asset]`
                    A list of file assets uploaded to the account.

            Return Fields

                created_at : `str`
                    The file's creation date.
                file_extension : `str`
                    The file's extension.
                file_size : `int`
                    The file's size in bytes.
                id : `str`
                    The file's unique identifier.
                name : `str`
                    The file's name.
                public_url : `str`
                    Public url to the asset, valid for 1 hour.
                uploaded_by : `moncli.entities.user.User`
                    The user who uploaded the file
                url : `str`
                    The user who uploaded the file
                url_thumbnail : `str`
                    Url to view the asset in thumbnail mode. Only available for images.   

            Optional Arguments

                column_ids : `list[str]`
                    A list of column IDs from which to retrieve file assets.     
        """
        
        kwargs = {'ids': [int(self.id)]}
        if column_ids:
            kwargs['assets'] = {'column_ids': column_ids}
        assets_data = api.get_items(
            *api.get_field_list(api.DEFAULT_ASSET_QUERY_FIELDS, 'assets', *args),
            api_key=self.__creds.api_key_v2,
            **kwargs)[0]['assets']
        return [en.Asset(**asset_data) for asset_data in assets_data]


    def remove_files(self, id: str = None, title: str = None, file_value=None, as_model: type = None, *args):
        """Removes a file from a column value.

            Parameters

                id : 'str'
                    The id of the column to be updated.
                title : 'str'
                    The title of the column to be updated.
                file_value : moncli.entities.column_value.FileValue
                    The file value to be removed.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional file return fields.

            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        
        if (id or title or file_value):
            if id:
                column_id = id
            if title:
                column_id = self.column_values[title].id
            if file_value:
                    column_id=file_value.id
        else:
            raise ItemError('clear_files_not_enough_parameters', self.id, 'Insufficient parameters for clearing files from column value.')

        item_data = api.change_column_value(
            item_id=self.id,
            board_id=self.board.id,
            column_id=column_id,
            value={'clear_all': True},
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]






    def get_board(self, *args):
        """Get the board that contains this item.

            Parameters

                args : `tuple`
                    Optional board return fields.

            Returns

                board : `moncli.entities.Board`
                    The board containing this item.

            Return Fields

                activity_logs : `list[moncli.entities.object.ActivityLog]`
                    The board log events.
                board_folder_id : `int`
                    The board's folder unique identifier.
                board_kind : `str`
                    The board's kind (public / private / share).
                columns : `list[moncli.entities.object.Column]`
                    The board's visible columns.
                communication : `str`
                    Get the board communication value - typically meeting ID.
                description : `str`
                    The board's description.
                groups : `list[moncli.entities.group.Group]`
                    The board's visible groups.
                id : `str`
                    The unique identifier of the board.
                items : `list[moncli.entities.item.Item]`
                    The board's items (rows).
                name : `str`
                    The board's name.
                owner : `moncli.entities.user.User`
                    The owner of the board.
                permissions : `str`
                    The board's permissions.
                pos : `str`
                    The board's position.
                state : `str`
                    The board's state (all / active / archived / deleted).
                subscribers : `list[moncli.entities.user.User]`
                    The board's subscribers.
                tags : `list[moncli.entities.objects.Tag]`
                    The board's specific tags.
                top_group : `moncli.entities.group.Group`
                    The top group at this board.
                updated_at : `str`
                    The last time the board was updated at (ISO8601 DateTime).
                updates : `list[moncli.entities.update.Update]`
                    The board's updates.
                views : `list[moncli.entities.board.BoardView]`
                    The board's views.
                workspace : `moncli.entities.objects.Workspace`
                    The workspace that contains this board (null for main workspace).
                workspace_id : `str`
                    The board's workspace unique identifier (null for main workspace).
        """

        board_data = api.get_items(
            *api.get_field_list(api.DEFAULT_BOARD_QUERY_FIELDS, 'board', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)])[0]['board']
        return en.Board(creds=self.__creds, **board_data)


    def get_group(self, *args):
        """Get the group that contains this item.

            Parameters

                args : `tuple`
                    Optional board return fields.

            Returns

                group : `moncli.entities.Group`
                    The group containing this item.


            Return Fields

                archived : `bool`
                    Is the group archived or not.
                color : `str`
                    The group's color.
                deleted : `bool`
                    Is the group deleted or not.
                id : `str`
                    The group's unique identifier.
                items : `list[moncli.entities.Item]`
                    The items in the group.
                position : `str`
                    The group's position in the board.
                title : `str`
                    The group's title.
        """

        group_data = api.get_items(
            *api.get_field_list(api.DEFAULT_GROUP_QUERY_FIELDS, 'group', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)])[0]['group']
        return en.Group(creds=self.__creds, __board=self, **group_data)


    def get_creator(self, *args):
        """Get the item's creator.

            Parameters

                args : `tuple`
                    The list of user return fields.

            Returns

                user : `moncli.entities.User`
                    The item's creator.

            Return Fields

                account : `moncli.entities.Account`
                    The user's account.
                birthday : `str`
                    The user's birthday.
                country_code : `str`
                    The user's country code.
                created_at : `str`
                    The user's creation date.
                email : `str`
                    The user's email.
                enabled : `bool`
                    Is the user enabled or not.
                id : `str`
                    The user's unique identifier.
                is_admin: `bool`
                    Is the user a admin or not.
                is_guest : `bool`
                    Is the user a guest or not.
                is_pending : `bool`
                    Is the user a pending user.
                is_verified: `bool`
                    Is the user is verified
                is_view_only : `bool`
                    Is the user a view only user or not.
                join_date : `str`
                    The date the user joined the account.
                location : `str`
                    The user' location.
                mobile_phone : `str`
                    The user's mobile phone number.
                name : `str`
                    The user's name.
                phone : `str`
                    The user's phone number.
                photo_original : `str`
                    The user's photo in the original size.
                photo_small : `str`
                    The user's photo in small size (150x150).
                photo_thumb : `str`
                    The user's photo in thumbnail size (100x100).
                photo_thumb_small : `str`
                    The user's photo in small thumbnail size (50x50).
                photo_tiny : `str`
                    The user's photo in tiny size (30x30).
                teams : `list[moncli.entities.Team]`
                    The teams the user is a member in.
                time_zone_identifier : `str`
                    The user's time zone identifier.
                title : `str`
                    The user's title.
                url : `str`
                    The user's profile url.
                utc_hours_diff : `int`
                    The user's UTC hours difference.
        """

        user_data = api.get_items(
            *api.get_field_list(api.DEFAULT_USER_QUERY_FIELDS, 'creator', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)])[0]['creator']
        return en.User(creds=self.__creds, **user_data)
   

    def get_column_values(self, *args):
        """Get the item's column values.

            Parameters

                args : `tuple`
                    Optional column value return fields.

            Returns

                column_value : `list[moncli.entities.ColumnValue]`
                    The item's column values.

            Return Fields

                additional_info : `json`
                    The column value's additional information.
                id : `str`
                    The column's unique identifier.
                text : `str`
                    The column's textual value in string form.
                title : `str`
                    The column's title.
                type : `str`
                    The column's type.
                value : `json`
                    The column's value in json format.
        """

        # Pulls the columns from the board containing the item and maps 
        # column ID to type.
        columns_map = { column.id: column for column in self.board.columns }
        column_values_data = api.get_items(
            *api.get_field_list(api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS, 'column_values', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)])[0]['column_values']

        values = []
        for data in column_values_data:
            id = data['id']
            column = columns_map[id]
            column_type = column.column_type
            data['settings_str'] = column.settings_str
            values.append(en.cv.create_column_value(column_type, **data))
        return en.BaseColumnCollection(values)


    def get_column_value(self, id = None, title = None, *args):
        """Get an item's column value by ID or title.

            Parameters

                id : `str`
                    The column's unique identifier.
                    NOTE: This parameter is mutually exclusive and cannot be used with 'title'.
                title : `str`
                    The column's title.
                    NOTE: This parameter is mutually exclusive and cannot be used with 'id'.
                args : `tuple`
                    Optional column value return fields.

            Returns

                column_value : `moncli.entities.ColumnValue`
                    The item's column value.

            Return Fields

                additional_info : `json`
                    The column value's additional information.
                id : `str`
                    The column's unique identifier.
                text : `str`
                    The column's textual value in string form.
                title : `str`
                    The column's title.
                type : `str`
                    The column's type.
                value : `json`
                    The column's value in json format.
        """

        if id and title:
            raise en.board.TooManyGetColumnValueParameters()
        if id is None and title is None:
            raise en.board.NotEnoughGetColumnValueParameters()
        
        if title:
            return self.column_values[title]

        return self.column_values[id]


    def get_subitems(self, *args):
        """Get the subitems of an item.

            Parameters

                args: list
                    list of return field arguments

            Returns

                subitems : `list[moncli.entities.Item]`
                    The item's nested subitems

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        subitems_data = api.get_items(
            *api.get_field_list(api.DEFAULT_ITEM_QUERY_FIELDS, 'subitems', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)])[0]['subitems']

        if not subitems_data:
            subitems_data = []
        return [en.Item(creds=self.__creds, **subitems) for subitems in subitems_data]


    def change_item_name(self, new_name: str,as_model: m.MondayModel = None):
        """Change the name of an item.

            Parameters

                nem_name : 'str'
                    The new name of the item.
                as_model: `type`
                    The MondayModel subclass to be returned.


            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        if new_name == None:
            raise InvalidParameterError()
        item_name = {'name':new_name}

        item_data = api.change_multiple_column_value(
            self.id,
            self.board.id,
            item_name,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]

    def change_column_value(self,id = None, title = None, column_value = None, get_column_values: bool = None, as_model: type = None, *args):
        """Get an item's column value by ID or title.

            Parameters
                id: str
                    The ID of the column to be updated.  
                    NOTE: This parameter is mutually exclusive and cannot be used with the 'title' parameter
                title: str
                    The title of the column value to be updated. 
                    NOTE: This parameter is mutually exclusive and cannot be used with 'id'.
                column_value : Type: moncli.entities.column_value.ColumnValue | dict | str
                    The value to be updated for the column.
                get_column_values: `bool`:
                    Retrieves item column values if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.

            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        if id and title : 
            raise e.ItemError(
                'change_column_value_too_many_parameters',
                self.id,
                'Cannot use both "id" and "title" parameters.'
            )
        if not (id or title):
            if isinstance(column_value, en.cv.ColumnValue):
                column_id = column_value.id
                value = column_value.format()
            else:
                raise e.ItemError(
                    'invalid_column_value',
                    self.id,
                    'Column value must be a properly formatted dict or str when using "id" or "title" parameters.'
                )
        elif id or title or column_value:
            value = None
            if isinstance(column_value,(dict,str)):
                if title:
                    column_id= self.column_values[title].id
                elif id:
                    column_id= id
                value = column_value
            if not column_value:
                if title:
                    column_value = self.column_values[title]
                elif id:
                    column_value = self.column_values[id]
                column_id = column_value.id
                value  = column_value.null_value
        else: 
            raise e.ItemError(
                'invalid_column_value_entity',
                self.id,
                'Column Value must be a valid entities.column_value.ColumnValue instance when not using "id" or "title" parameters.'
            )

        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])


        item_data = api.change_column_value(
            self.id,
            column_id,
            self.board.id,
            value,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]

    def change_simple_column_value(self, id = None, title = None, value = None,as_model: type = None, *args):
        """Change the item's column values using simple values.

            Parameters

                id: str
                    The id value of the column 
                title: str 
                    The title of the column
                as_model: type
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.

            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        
        if not id and not title :
            raise NotEnoughChangeSimpleColumnValueParameters()
        if id and title :
            raise TooManyChangeSimpleColumnValueParameters()
                
        if id:
            column_value = self.column_values[id]   
        elif title:
            column_value = self.column_values[title]
        
        item_data = api.change_simple_column_value(
            self.id,
            self.board.id,
            column_value.id,
            value,
            *args,
            api_key=self.__creds.api_key_v2)        

        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]

    def change_multiple_column_values(self, column_values, get_column_values: bool = False, as_model: type = None, *args):
        """Change the item's column values.

            Parameters

                column_values : `list[moncli.entities.ColumnValue] / dict`
                    The column value to update. 
                    NOTE: This value can either be a list of moncli.entities.ColumnValue objects or a formatted dictionary.
                get_column_values: `bool`:
                    Retrieves item column values if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.

            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """

        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])

        if type(column_values) == dict:
            values = column_values
        elif type(column_values) == list:
            values = { value.id: value.format() for value in column_values }
        else:
            raise en.InvalidColumnValue(type(column_values).__name__)
        item_data = api.change_multiple_column_value(
            self.id,
            self.board.id,
            values,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


    def create_subitem(self, item_name: str, as_model: type = None, *args, **kwargs):
        """Create subitem.

            Parameters
            
                item_name : `str`
                    The new item's name.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item return fields.
                kwargs : `dict`
                    Optional arguments for creating subitems.

            Returns
                        
                subitem : `moncli.entities.Item`
                    The created subitem.

            Return Fields
            
                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments
            
                column_values : `json`
                    The column values of the new item.
        """
        
        subitem_data = api.create_subitem(
            self.id,
            item_name,
            *args,
            api_key=self.__creds.api_key_v2,
            **kwargs)
        items = Item(creds=self.__creds, **subitem_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]



    def move_to_group(self, group_id: str, get_column_values = False, as_model: type = None, *args):
        """Move item to a different group.

            Parameters

                group_id : `str`
                    The group's unique identifier.
                get_column_values: `bool`:
                    Retrieves item column values if set to `True`.
                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.

            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """

        if get_column_values:
            args = list(args)
            for arg in ['column_values.{}'.format(arg) for arg in api.DEFAULT_COLUMN_VALUE_QUERY_FIELDS]:
                if arg not in args:
                    args.append(arg)
            args.extend(['id', 'name'])

        item_data = api.move_item_to_group(
            self.id,
            group_id,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]



    def archive(self, as_model: type = None, *args):
        """Archive this item.

            Parameters

                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.


            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """

        item_data = api.archive_item(
            self.id,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]



    def delete(self, as_model: type = None, *args):
        """Delete this item.

            Parameters

                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    Optional item return fields.


            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """

        item_data = api.delete_item(
            self.id,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]



    def duplicate(self, as_model: type = None, *args, **kwargs):
        """Duplicate this item.

            Parameters

                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of item return fields.
                kwargs : `dict`
                    Optional keyword arguments for duplicating item.

            Returns
                
                item : `moncli.entities.Item`
                    The duplicated item.

            Return Fields
            
                assets : `list[moncli.entities.Asset]`
                    The item's assets/files.
                board : `moncli.entities.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.Update`
                    The item's updates.
            
            Optional Arguments

                with_updates : `bool`
                    Duplicate with the item's updates.
        """

        item_data = api.duplicate_item(
            self.board.id,
            self.id,
            *args,
            api_key=self.__creds.api_key_v2,
            *kwargs)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]



    def add_update(self, body: str, *args, **kwargs):
        """Change the item's column values.

            Parameters

                body : `str`
                    The update text.
                args : `tuple`
                    Optional update return fields.
                kwargs : `dict`
                    Optional keyword arguments for adding an update.

            Returns

                update : `moncli.entities.Update`
                    The created update.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.reply.Reply]
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.

            Optional Arguments

                parent_id : `str`
                    The parent post identifier.
        """

        update_data = api.create_update(
            body, 
            self.id,
            *args,
            api_key=self.__creds.api_key_v2, 
            **kwargs)
        return en.Update(creds=self.__creds, **update_data)


    def get_updates(self, *args, **kwargs):
        """Get updates for this item.
 
            Parameters
    
                args : `tuple`
                    Optional update return fields.
                kwargs : `dict`
                    Optional keyword arguments for getting item updates.

            Returns

                update : `list[moncli.entities.Update]`
                    The item's updates.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.entities.Reply]`
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.

            Optional Arguments

                limit : `int`
                    Number of updates to get; the default is 25.
                page : `int`
                    Page number to get, starting at 1.
        """
        
        limit = kwargs.pop('limit', 25)
        page = kwargs.pop('page', 1)
        updates_data = api.get_items(
            *api.get_field_list(api.DEFAULT_UPDATE_QUERY_FIELDS, 'updates', *args),
            api_key=self.__creds.api_key_v2,
            ids=[int(self.id)],
            limit=1,
            updates={'limit': limit, 'page': page})[0]['updates']
        return [en.Update(creds=self.__creds, **update_data) for update_data in updates_data]


    def delete_update(self, update_id: str, *args):
        """Delete an item's update

            Parameters

                update_id : `str`
                    The update's unique identifier
                args : `tuple`
                    The list of optional fields to return.

            Returns

                update : `moncli.entities.Update`
                    The item's deleted update.

            Return Fields

                assets : `list[moncli.entities.Asset]`
                    The update's assets/files.
                body: `str`
                    The update's html formatted body.
                created_at: `str`
                    The update's creation date.
                creator : `moncli.entities.User`
                    The update's creator
                creator_id : `str`
                    The unique identifier of the update creator.
                id : `str`
                    The update's unique identifier.
                item_id : `str`
                    The update's item ID.
                replies : `list[moncli.entities.Reply]
                    The update's replies.
                text_body : `str`
                    The update's text body.
                updated_at : `str`
                    The update's last edit date.
        """

        updates = self.get_updates(*args)
        target_update = [update for update in updates if update.id == update_id]
        if not target_update:
            raise UpdateNotFound(update_id)
        return target_update[0].delete()


    def clear_updates(self, as_model: type = None, *args):
        """Clear item's updates.

            Parameters

                as_model: `type`
                    The MondayModel subclass to be returned.
                args : `tuple`
                    The list of optional fields to return.


            Returns

                item : `moncli.entities.Item`
                    The updated item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """

        item_data = api.clear_item_updates(
            self.id,
            *args,
            api_key=self.__creds.api_key_v2)
        items = Item(creds=self.__creds, **item_data)
        if not as_model:
            return items
        if not issubclass(type(as_model), MondayModel):
            raise ItemError(
                'invalid_as_model_parameter',
                self.id,
                'as_model parameter must be of MondayModel Type')
        return [as_model(item) for item in items]


    
    def get_activity_logs(self, *args, **kwargs):
        """Retrieves an item's activity logs.

            Parameters
            
                args : `tuple`
                    The list of activity log return fields.
                kwargs : `dict`
                    Optional keyword arguments for retrieving activity logs.

            Returns

                activity_logs : `list[moncli.entities.ActivityLog]`
                    The board's activity logs.

            Return Fields

                account_id : `str`
                    The unique identifier of the user's account.
                created_at : `str`
                    The create date
                data : `str`
                    The item's column values in string form.
                entity : `str`
                    The activity log's entity.
                event : `str`
                    The activity log's event.
                id : `str`
                    The activity log's unique identifier.
                user_id : `str`
                    The user's unique identifier.

            Optional Arguments

                limit : `int`
                    Number of items to get, the default is 25.
                page : `int`
                    Page number to get, starting at 1.
                user_ids : `list[str]`
                    User ids to filter.
                column_ids : `list[str]`
                    Column ids to filter.
                group_ids : `list[str]`
                    Group ids to filter.
                from : `str`
                    From timespamp (ISO8601).
                to : `str`
                    To timespamp (ISO8601).
        """

        
        kwargs['item_ids'] = [int(self.id)]
        return self.board.get_activity_logs(*args, **kwargs)

    def get_parent_item(self, *args, **kwargs):
        """Get Parent Item of a Subitem.

            Parameters

                args : `tuple`
                    Optional item return fields.

            Returns

                item : `moncli.entities.Item`
                    The parent item.

            Return Fields

                assets : `list[moncli.entities.asset.Asset]`
                    The item's assets/files.
                board : `moncli.entities.board.Board`
                    The board that contains this item.
                column_values : `list[moncli.entities.column_value.ColumnValue]`
                    The item's column values.
                created_at : `str`
                    The item's create date.
                creator : `moncli.entities.user.User`
                    The item's creator.
                creator_id : `str`
                    The item's unique identifier.
                group : `moncli.entities.group.Group`
                    The group that contains this item.
                id : `str`
                    The item's unique identifier.
                name : `str`
                    The item's name.
                state : `str`
                    The board's state (all / active / archived / deleted)
                subscriber : `moncli.entities.user.User`
                    The pulse's subscribers.
                updated_at : `str`
                    The item's last update date.
                updates : `moncli.entities.update.Update`
                    The item's updates.
        """
        item_data = api.get_items(
            *api.get_field_list(api.DEFAULT_ITEM_QUERY_FIELDS, 'parent_item', *args),
            ids=[self.id]
        )[0]['parent_item']
        return Item(creds=self.__creds, **item_data)

class ColumnValueRequired(Exception):
    def __init__(self):
        self.message = "A column value is required if no 'column_id' value is present."

class UpdateNotFound(Exception):
    def __init__(self, update_id: str):
        self.message = "Item does not contain update with ID '{}'.".format(update_id)

class TooManyChangeSimpleColumnValueParameters(Exception):
    def __init__(self):
        self.message = "Unable to use both 'id' and 'title' when changing a simple column value"

class NotEnoughChangeSimpleColumnValueParameters(Exception):
    def __init__(self):
        self.message = "Either 'id' or 'title' is required when changing a simple column value"

class InvalidParameterError(Exception):
    def __init__(self):
        self.message = "New name must be present"




BASE_URL = 'https://api.monday.com'
PORT = '443'
API_VERSION = 'v1'

# User route signatures
USERS = 'users.json'
USER_BY_ID = 'users/{}.json'
USER_POSTS = 'users/{}/posts.json'
USER_NEWSFEED = 'users/{}/newsfeed.json'
USER_UNREAD_FEED = 'users/{}/unread_feed'

# Updates route signatures
UPDATES = 'updates.json'
UPDATES_BY_ID = 'updates/{}.json'
UPDATE_LIKE = 'updates/{}/like.json'
UPDATE_UNLIKE = 'updates/{}/unlike.json'

# Pulses route signatures
PULSES = 'pulses.json'
PULSE_BY_ID = 'pulses/{}.json'
PULSE_SUBSCRIBERS = 'pulses/{}/subscribers.json'
PULSE_SUBSCRIBERS_BY_USER_ID = 'pulses/{}/subscribers/{}.json'
PULSE_NOTES = 'pulses/{}/notes.json'
PULSE_NOTE_BY_ID = 'pulses/{}/notes/{}.json'
PULSE_UPDATES = 'pulses/{}/updates.json'

# Boards route signatures
BOARDS = 'boards.json'
BOARD_BY_ID = 'boards/{}.json'
BOARD_GROUPS = 'boards/{}/groups.json'
BOARD_MOVE_GROUP = 'boards/{}/groups/{}/move.json'
BOARD_GROUP_BY_ID = 'boards/{}/groups/{}.json'
BOARD_COLUMNS = 'boards/{}/columns.json'
BOARD_COLUMN_BY_ID = 'boards/{}/columns/{}.json'
BOARD_COLUMN_VALUE = 'boards/{}/columns/{}/value.json'
BOARD_TEXT_COLUMN = 'boards/{}/columns/{}/text.json'
BOARD_PERSON_COLUMN = 'boards/{}/columns/{}/person.json'
BOARD_STATUS_COLUMN = 'boards/{}/columns/{}/status.json'
BOARD_DATE_COLUMN = 'boards/{}/columns/{}/date.json'
BOARD_NUMERIC_COLUMN = 'boards/{}/columns/{}/numeric.json'
BOARD_TAGS_COLUMN = 'boards/{}/columns/{}/tags.json'
BOARD_TIMELINE_COLUMN = 'boards/{}/columns/{}/timeline.json'
BOARD_PULSES = 'boards/{}/pulses.json'
BOARD_DUPLICATE_PULSE = 'boards/{}/pulses/{}/duplicate.json'
BOARD_MOVE_PULSES = 'boards/{}/pulses/move.json'
BOARD_SUBSCRIBERS = 'boards/{}/subscribers.json'
BOARD_SUBSCRIBERS_BY_ID = 'boards/{}/subscribers/{}.json'

# Tags route signatures
TAGS_BY_ID = 'tags/{}.json'


from moncli.routes import constants
from moncli.routes.requests import execute_get, execute_post, execute_delete

def get_updates(
        api_key, 
        page = 1, 
        per_page = 25, 
        offset = 0, 
        since = None, 
        until = None, 
        updated_since = None, 
        updated_until = None):

    resource_url = constants.UPDATES

    params = {
        'page': page,
        'per_page': per_page,
        'offset': offset
    }

    if since != None:
        params['since'] = since

    if since != None:
        params['until'] = until

    if since != None:
        params['updated_since'] = updated_since

    if since != None:
        params['updated_until'] = updated_until

    return execute_get(api_key, resource_url, params)


def post_update(api_key, user_id, pulse_id, update_text):

    resource_url = constants.BOARDS

    body = {
        'user_id': user_id,
        'pulse_id': pulse_id,
        'update_text': update_text
    }

    return execute_post(api_key, resource_url, body)


def get_update_by_id(api_key, update_id):

    resource_url = constants.UPDATES_BY_ID.format(update_id)

    return execute_get(api_key, resource_url)


def delete_update_by_id(api_key, update_id):

    resource_url = constants.UPDATES_BY_ID.format(update_id)

    return execute_delete(api_key, resource_url)


def post_like_to_update(api_key, update_id, user_id):

    resource_url = constants.UPDATE_LIKE.format(update_id)

    body = {
        'user': user_id
    }

    return execute_post(api_key, resource_url, body)


def post_unlike_to_update(api_key, update_id, user_id):

    resource_url = constants.UPDATE_UNLIKE.format(update_id)

    body = {
        'user': user_id
    }

    return execute_post(api_key, resource_url, body)
